base_urls = {
    "openai" : "https://api.openai.com/v1/",
    "deepseek" : "https://api.deepseek.com/v1/",
    "llama.cpp" : "http://localhost:8080/v1/",
    "openrouter" : "https://openrouter.ai/api/v1/"
}
api_keys = {
    "llama.cpp" : "sk-no-key-required"
}
models = {
    "openai" : "gpt-4o",
    "deepseek" : "deepseek-chat",
    "llama.cpp" : ""
}
